
/*!
 * \file
 *
 * \brief GenlibClientSubscription object implementation.
 *
 * \author Marcelo Roberto Jimenez
 */

#include "config.h"

#ifdef INCLUDE_CLIENT_APIS

#define TEMPLATE_GENERATE_SOURCE
#include "ClientSubscription.h"

#endif /* INCLUDE_CLIENT_APIS */

