
/*!
 * \file
 *
 * \brief UpnpSubscriptionRequest object implementation.
 *
 * \author Marcelo Roberto Jimenez
 */

#include "config.h"

#define TEMPLATE_GENERATE_SOURCE
#include "SubscriptionRequest.h"

